import typing

from mccq.node.abc.node import Node
from mccq.node.data_node import DataNode


class QueryNode(Node):
    def __init__(self, data_node: DataNode, children: typing.Tuple['QueryNode', ...] = None):
        self.data_node = data_node
        self._children = children

    def leaves(self) -> typing.Iterable['QueryNode']:
        return super().leaves()

    @property
    def children(self) -> typing.Tuple['QueryNode', ...]:
        return self._children or ()



import logging
import typing
import urllib.parse

from mccq import errors
from mccq.data_loader.abc.data_loader import DataLoader
from mccq.data_loader.filesystem_data_loader import FilesystemDataLoader
from mccq.data_loader.internet_data_loader import InternetDataLoader
from mccq.data_parser.abc.data_parser import DataParser
from mccq.data_parser.v1_data_parser import V1DataParser
from mccq.node.data_node import DataNode
from mccq.typedefs import IterableOfStrings, TupleOfStrings

log = logging.getLogger(__name__)

LoaderGeneric = typing.Union[str, DataLoader]
ParserGeneric = typing.Union[str, DataParser]

LOADER_MAP = {
    'file': FilesystemDataLoader,
    'http': InternetDataLoader,
    'https': InternetDataLoader
}

PARSER_MAP = {
    'v1': V1DataParser
}

DATA_FILE_TAIL = ('generated', 'reports', 'commands.json')


def find_loader(obj, uri) -> DataLoader:
    try:
        if obj is None:
            # auto-detect database source to instantiate an appropriate loader
            uri_scheme = urllib.parse.urlparse(uri).scheme
            return LOADER_MAP.get(uri_scheme, LOADER_MAP['file'])()

        elif isinstance(obj, str):
            return LOADER_MAP[obj]()

        elif isinstance(obj, DataLoader):
            return obj

    except Exception as ex:
        raise errors.InvalidLoader(obj) from ex


def find_parser(obj) -> DataParser:
    try:
        if obj is None:
            # default to the only available parser
            return PARSER_MAP['v1']()

        elif isinstance(obj, str):
            return PARSER_MAP[obj]()

        elif isinstance(obj, DataParser):
            return obj

    except Exception as ex:
        raise errors.InvalidParser(obj) from ex


class VersionDatabase:
    def __init__(
            self, uri: str, loader: LoaderGeneric = None, parser: ParserGeneric = None, version_file: str = None,
            whitelist: IterableOfStrings = ()):
        self.uri = uri
        self.version_file = version_file
        self.whitelist = set(whitelist)
        self.loader: DataLoader = find_loader(loader, uri)
        self.parser: DataParser = find_parser(parser)
        self._node_cache: typing.Dict[str, DataNode] = {}
        self._version_cache: typing.Dict[str, str] = {}

    def _load_actual_version(self, version: str) -> str:
        return self.loader.load_version((self.uri, version, self.version_file))

    def _load(self, version: str):
        components = (self.uri, version, *DATA_FILE_TAIL)

        try:
            actual_version = self._load_actual_version(version)
            self._version_cache[version] = actual_version
            log.info(f'Loading commands for version {version} (actual {actual_version}) with components: {components}')
        except:
            log.info(f'Loading commands for version {version} with components: {components}')

        # load data from source
        try:
            raw = self.loader.load(components)
        except Exception as ex:
            raise errors.LoaderFailure(version) from ex

        # parse and insert data
        try:
            parsed = self.parser.parse(raw)
        except Exception as ex:
            raise errors.ParserFailure(version) from ex

        self.put(version, parsed)

    def get_actual_version(self, version: str) -> str:
        return self._version_cache.get(version)

    def reload(self):
        self._node_cache = {}
        self._version_cache = {}

    def get(self, version: str) -> DataNode:
        log.debug(f'Getting root node for version {version}')
        if version not in self._node_cache:
            log.info(f'Loading version {version} into cache')
            self._load(version)
        return self._node_cache[version]

    def put(self, version: str, root_node: DataNode):
        if self.whitelist and version not in self.whitelist:
            raise errors.VersionNotWhitelisted(version)
        self._node_cache[version] = root_node

    def filter_versions(self, requested_versions: IterableOfStrings) -> TupleOfStrings:
        # if no whitelist, everything is valid
        if not self.whitelist:
            return requested_versions
        # otherwise filter out versions that are not whitelisted
        available_versions = set(requested_versions).intersection(self.whitelist)
        # make sure to preserve order
        return tuple(version for version in requested_versions if version in available_versions)



from mccq.node.data_node import DataNode
from mccq.data_parser.abc.data_parser import DataParser


class V1DataParser(DataParser):
    def _build(self, key: str, node: dict, command: str, command_t: str):
        type_ = node['type']
        executable = node.get('executable')
        redirect = node.get('redirect')
        children = node.get('children', {})

        # whether my command is relevant enough to be rendered
        relevant = executable

        # build my argument string
        if type_ == 'root':
            args = args_t = ()
        elif type_ == 'literal':
            args = args_t = (key,)
        elif type_ == 'argument':
            parser = node['parser'].split(sep=':', maxsplit=1)[1]  # get the `string` from `brigadier:string`
            args = ('<{}>'.format(key),)
            args_t = ('<{}: {}>'.format(key, parser),)
        else:
            args = args_t = ('{}*'.format(key),)

        # argument to provide for parents when collapsing
        argument = args[0] if args else None
        argument_t = args_t[0] if args_t else None

        if redirect:
            # redirect is a list and there may be multiple
            args += ('->', '|'.join(redirect))
            relevant = True

        # special case for `execute run`
        if not (executable or redirect or children):
            args += ('->', '*')
            relevant = True

        # build command
        my_command = ' '.join(arg for arg in ((command or None,) + args) if arg is not None)
        my_command_t = ' '.join(arg_t for arg_t in ((command_t or None,) + args_t) if arg_t is not None)

        # build children, if any
        my_children = tuple(self._build(k, v, my_command, my_command_t) for k, v in children.items())

        # count population
        population = sum(child.population for child in my_children)
        if relevant:
            population += 1

        # build collapsed form
        collapsed = collapsed_t = None
        if my_children:
            collapsed = ' '.join((my_command, '|'.join((child.argument for child in my_children))))
            collapsed_t = ' '.join((my_command_t, '|'.join((child.argument_t for child in my_children))))
            # look for at least one grandchild before appending `...`
            if next((True for child in my_children if child.children), False):
                collapsed += ' ...'
                collapsed_t += ' ...'

        return DataNode(
            relevant=relevant,
            population=population,
            key=key,
            command=my_command,
            command_t=my_command_t,
            argument=argument,
            argument_t=argument_t,
            collapsed=collapsed,
            collapsed_t=collapsed_t,
            children=my_children,
        )

    def parse(self, raw) -> DataNode:
        return self._build('root', raw, '', '')



class MCCQError(Exception):
    """ MCCQ base error. """


class ArgumentParserFailed(MCCQError):
    """ Raised when the argument parser fails to process the given command string. """
    def __init__(self, command: str, *args):
        super().__init__(*args)
        self.command = command

    def __str__(self):
        return f'Argument parsers failed to process command: {self.command}'


class NoVersionRequested(MCCQError):
    """ Raised when no versions were requested. """

    def __str__(self):
        return 'No versions were requested'


class NoVersionsAvailable(MCCQError):
    """ Raised when none of the requested versions are available. """

    def __init__(self, requested_versions: tuple, *args):
        super().__init__(*args)
        self.versions = requested_versions

    def __str__(self):
        versions_str = ', '.join(self.versions)
        return f'None of the requested versions are available: {versions_str}'


class NoSuchVersion(MCCQError):
    """ Raised when the accessed version is not available. """
    def __init__(self, version: str, *args):
        super().__init__(*args)
        self.version = version

    def __str__(self):
        return f'Version {self.version} is not available'


class VersionNotWhitelisted(MCCQError):
    """ Raised when the version whitelist is enabled and the requested version is not present. """
    def __init__(self, version: str, *args):
        super().__init__(*args)
        self.version = version

    def __str__(self):
        return f'Version {self.version} is not whitelisted'


class MissingCommand(MCCQError):
    """ Raised when the provided command is empty or null. """

    def __str__(self):
        return 'No command was provided'


class NoSuchCommand(MCCQError):
    """ Raised when the given command does not contain a valid base command. """

    def __init__(self, command: str, *args):
        super().__init__(*args)
        self.command = command

    def __str__(self):
        return f'Command does not exist: {self.command}'


class InvalidLoader(MCCQError):
    """ Raised when an invalid data parser is provided. """

    def __init__(self, loader: str, *args):
        super().__init__(*args)
        self.loader = loader

    def __str__(self):
        return f'Invalid data loader: {self.loader}'


class LoaderFailure(MCCQError):
    """ Raised when the data file for the given version failed to load. """

    def __init__(self, version: str, *args):
        super().__init__(*args)
        self.version = version

    def __str__(self):
        return f'Failed to load data for version {self.version}'


class InvalidParser(MCCQError):
    """ Raised when an invalid data parser is provided. """

    def __init__(self, parser: str, *args):
        super().__init__(*args)
        self.parser = parser

    def __str__(self):
        return f'Invalid data parser: {self.parser}'


class ParserFailure(MCCQError):
    """ Raised when a parser fails to process data. """

    def __init__(self, version: str, *args):
        super().__init__(*args)
        self.version = version

    def __str__(self):
        return f'Failed to parse commands for version {self.version}'


import json
import logging
import urllib.request

from mccq.data_loader.abc.data_loader import DataLoader
from mccq.typedefs import TupleOfStrings

log = logging.getLogger(__name__)


class InternetDataLoader(DataLoader):
    def load(self, components: TupleOfStrings) -> dict:
        path = '/'.join(components)
        log.info(f'Loading commands from internet: {path}')
        response = urllib.request.urlopen(path)
        content = response.read().decode('utf8')
        raw = json.loads(content)
        return raw

    def load_version(self, components: TupleOfStrings) -> str:
        path = '/'.join(components)
        log.info(f'Loading version from internet: {path}')
        response = urllib.request.urlopen(path)
        content = response.readline().decode('utf8')
        raw = str(content).strip()
        return raw



import logging
import shlex

from mccq import errors
from mccq.cli.meta import META_MAP
from mccq.query_manager import QueryManager

log = logging.getLogger(__name__)


def cli_loop(qm: QueryManager):
    while True:
        try:
            command = input('> ')

        except KeyboardInterrupt:
            print()
            return

        if command.startswith('\\'):
            try:
                meta_command = command[1:]
                meta_args = shlex.split(meta_command)
                meta_root = meta_args[0]

                if meta_root in META_MAP['exit']:
                    return

                elif meta_root in META_MAP['reload']:
                    qm.database.reload()

                elif meta_root in META_MAP['show']:
                    qm.show_versions = tuple(meta_args[1:])

                else:
                    raise ValueError('Invalid command', command)

            except Exception:
                print('Invalid command, please try again.')
                if log.isEnabledFor(logging.DEBUG):
                    log.exception('Error')

        elif command:
            try:
                for version, commands in qm.results(command).items():
                    print(f'# {version}')
                    for command in commands:
                        print(command)

            except errors.NoVersionRequested:
                print('No versions provided, use \\s to set the default(s).')

            except Exception as ex:
                print(f'Error: {ex}')
                if log.isEnabledFor(logging.DEBUG):
                    log.exception('Error')



import re
import shlex
import typing

from mccq import errors
from mccq.argument_parser import ArgumentParser
from mccq.node.data_node import DataNode
from mccq.node.query_node import QueryNode
from mccq.query_arguments import QueryArguments
from mccq.typedefs import IterableOfStrings, TupleOfStrings
from mccq.version_database import VersionDatabase

# map of version names to command results
# example: `{'18w01a': ('tag <targets> add <tag>', 'tag <targets> remove <tag>')}`
QueryResults = typing.Dict[str, TupleOfStrings]


class QueryManager:
    ARGUMENT_PARSER = ArgumentParser(
        'mccq',
        description='Minecraft command query program. Inspired by the in-game help command, with added features like '
                    'version reporting and expandable regex search.',
        add_help=False)

    ARGUMENT_PARSER.add_argument(
        '-t', '--showtypes', action='store_true', help='whether to show argument types')

    ARGUMENT_PARSER.add_argument(
        '-e', '--explode', action='store_true', help='whether to expand all subcommands, regardless of capacity')

    ARGUMENT_PARSER.add_argument(
        '-c', '--capacity', type=int, default=12, help='maximum number of subcommands to render before collapsing')

    ARGUMENT_PARSER.add_argument(
        '-v', '--version', action='append', default=[], help='which version(s) to use for the command (repeatable)')

    ARGUMENT_PARSER.add_argument(
        'command', nargs='+', help='the command query')

    def __init__(
            self,
            database: VersionDatabase,
            show_versions: IterableOfStrings,
    ):
        self.database = database
        self.show_versions: TupleOfStrings = tuple(show_versions)

    @staticmethod
    def parse_query_arguments(command: str) -> QueryArguments:
        try:
            # split into tokens using shell-like syntax (preserve quoted substrings)
            parsed_args = QueryManager.ARGUMENT_PARSER.parse_args(shlex.split(command))

            # return an object representation
            return QueryArguments(
                command=tuple(parsed_args.command),  # immutable copy
                showtypes=parsed_args.showtypes,
                explode=parsed_args.explode,
                capacity=parsed_args.capacity,
                versions=tuple(parsed_args.version),  # duplicate versions are meaningless
            )

        except Exception as ex:
            raise errors.ArgumentParserFailed(command) from ex

    def _query_tree_recursive(self, arguments: QueryArguments, node: DataNode, index: int) \
            -> typing.Union[QueryNode, None]:
        # determine the current search term
        token = arguments.command[index] if len(arguments.command) > index else None

        # use regex to search for the subcommand/argument name in the patternized token
        # special case: dot matches all
        search_children = None if not token else node.children if token in ('.', '*') else tuple(
            child for child in node.children
            if re.search(token, child.key, re.IGNORECASE)
        )

        # branch: search all matching children recursively (depth-first) for subcommands
        if search_children:
            query_children = tuple(item for item in (
                self._query_tree_recursive(arguments, child, index + 1) for child in search_children
            ) if item is not None)

            if query_children:
                # return a query node with the calculated subset of children
                return QueryNode(data_node=node, children=query_children)

        # leaf: no children to search and tokens depleted; return a childless query node
        # note that even though this is a recursive leaf, the contained data node may itself have children
        elif not token:
            return QueryNode(data_node=node)

        # at this point 'else' means there are still tokens to search, so the query goes deeper than the current node
        # and we can just ignore it

    def _commands_recursives(self, arguments: QueryArguments, node: DataNode) -> IterableOfStrings:
        command = node.command_t if arguments.showtypes else node.command
        collapsed = node.collapsed_t if arguments.showtypes else node.collapsed

        # render relevant commands:
        #   - all executable commands: `scoreboard players list`, `scoreboard players list <target>`
        #   - all chainable (redirect) commands: `execute as <entity> -> execute`
        #   - some exceptional cases: `execute run ...`
        if node.relevant:
            yield command

        # determine whether to continue searching any existing children for subcommands
        # if any of the following are true, continue searching:
        #   1. explode override flag is set
        #   2. capacity has not been reached
        #   3. only one child to search
        if node.children and (
                arguments.explode
                or (node.population <= arguments.capacity)
                or len(node.children) == 1
        ):
            for child in node.children:
                yield from self._commands_recursives(arguments, child)

        # otherwise render a collapsed form
        elif collapsed:
            yield collapsed

    def filter_versions(self, arguments: QueryArguments) -> TupleOfStrings:
        requested_versions = arguments.versions or self.show_versions

        # make sure at least one version was requested
        if not requested_versions:
            raise errors.NoVersionRequested()

        # filter out unavailable versions
        filtered_versions = self.database.filter_versions(requested_versions)

        # make sure at least one of the requested versions is available
        if not filtered_versions:
            raise errors.NoVersionsAvailable(requested_versions)

        return filtered_versions

    def query_tree_for_version(self, version: str, arguments: QueryArguments) -> typing.Union[QueryNode, None]:
        # get the root data node and make sure the version is loaded
        root_data_node = self.database.get(version)
        if not root_data_node:
            raise errors.NoSuchVersion(version)

        # build a trimmed tree containing only the nodes that match the given arguments
        query_tree = self._query_tree_recursive(arguments, root_data_node, index=0)

        return query_tree

    def commands_for_version(self, version: str, arguments: QueryArguments) -> IterableOfStrings:
        # first build a result tree from the given arguments
        query_tree = self.query_tree_for_version(version, arguments)

        # proceed only if the query returned something
        if query_tree:
            # then check every branch and leaf for relevant commands
            # get all leaves of the query tree, and then all of the leaves of their corresponding data nodes
            for leaf in query_tree.leaves():
                yield from self._commands_recursives(arguments, leaf.data_node)

    def results_from_versions(self, versions: IterableOfStrings, arguments: QueryArguments) -> QueryResults:
        # ignore errors when multiple versions are specified
        # (not sure how else to handle this gracefully)
        results = {}
        for version in versions:
            try:
                commands = tuple(self.commands_for_version(version, arguments))

            # ignore errors because we may have other results
            except:
                continue

            # don't include versions with no results
            if commands:
                results[version] = commands

        return results

    def results_from_version(self, version: str, arguments: QueryArguments) -> QueryResults:
        # handle single version requests differently by allowing errors to propagate
        commands = tuple(self.commands_for_version(version, arguments))
        return {version: commands} if commands else {}

    def results_from_arguments(self, arguments: QueryArguments) -> QueryResults:
        filtered_versions = self.filter_versions(arguments)
        if len(filtered_versions) > 1:
            results = self.results_from_versions(filtered_versions, arguments)
        else:
            results = self.results_from_version(filtered_versions[0], arguments)
        return results

    def results(self, command: str) -> QueryResults:
        return self.results_from_arguments(self.parse_query_arguments(command))

    def reload(self):
        self.database.reload()
